//-------------------------------------------------------
// Coefficients of One-dimensional N-Atom Network (CONAN)
//-------------------------------------------------------

//    Copyright (C) 2016  The Few-Body Group, Department for Physics
//    and Astronomy, Aarhus University, Denmark

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cvector.h"
#include "expparser.h"
#include "physics.h"
#include "output.h"
#include <time.h>
#include <getopt.h>

//-------------------------------
// Forward declarations
//-------------------------------

struct expression;
struct function;

// Solve the actual problem
void doSolve(int dim, struct result* r, int calculate_coeffecients)
{
    gsl_matrix* C = gsl_matrix_alloc(dim, basesize);
    gsl_vector* E = gsl_vector_alloc(basesize);
    Solver(C, E);
    gsl_vector* Ik_vec = gsl_vector_alloc(basesize);
    if(calculate_coeffecients) {
#pragma omp parallel for
        for (int k = 1; k < dim; k++) {
            double Ik = I(dim, k, E, C);
            gsl_vector_set(Ik_vec, k-1, Ik);
        }
    }

    gsl_matrix_free(C);
    r->E = E;
    r->Ik = Ik_vec;
}

// Struct to hold an input parameter. It can hold a name and an array
// of values.
typedef struct {
    int n;
    char name[20];
    double* array;
} input_parameter;

void parse_parameter(const char* arg, input_parameter* result)
{
    size_t eqpos = strcspn(arg, "=");
    if (eqpos == 0 || arg[eqpos] != '=') {
        fprintf(stderr, "Invalid variable assignment: %s\n", arg);
        exit(1);
    }
    strncpy(result->name, arg, eqpos);
    char* first_paranthesis = strchr(arg, '[');
    // It is simple one value and we can add it normally
    if (!first_paranthesis) {
        result->array = (double*) malloc(sizeof(double));
        result->array[0] = strtod(arg + eqpos + 1, NULL);
        result->n = 1;
        return;
    }
    // It is an array and we need to parse each value
    // Count number of commas, and infer number of variables
    int n;
    char* s = first_paranthesis;
    for (n=0; s[n]; s[n] == ',' ? n++ : *s++);
    n++; // One more number than commas

    result->array = (double*) malloc(n*sizeof(double));

    for(int i=0; i<n; i++) {
        result->array[i] = strtod(first_paranthesis + 1, &first_paranthesis);
    }
    result->n = n;
}

void printLongUsage() {
    printf("Usage: Conan [-V/--potential] [-N/--number-of-par] [-options] [<params>]\n");
    printf("             -L/--wellwidth\n");
    printf("             -b/--basesize\n");
    printf("             -p/--precision\n");
    printf("             --abs-solver\n");
    printf("             --abs-solver\n");
    printf("             --rel-solver\n");
    printf("             --abs-final\n");
    printf("             --rel-final\n");
    printf("             -D/--just-potential\n");
    printf("             -E/--just-energy\n");
    printf("             -o/--filename\n");
}

int is_number(const char* input) {
    for(size_t i = 0; input[i] != '\0'; ++i) {
        if(! (isdigit(input[i]) || '.' == input[i]) ) {
            return 0;
        }
    }
    return 1;
}

void sanity_check(char** argv, int N, int prec, double wellwidth, int basesize, char* pot) {
    if (N==-1) {
        printLongUsage();
        exit(1);
    }
    if (pot == NULL) {
        printLongUsage();
        exit(1);
    }
    if (N <= 1) {
        fprintf(stderr, "%s: N=%i is too small\n", argv[0], N);
        exit(1);
    }

    if (prec < MPFR_PREC_MIN) {
        fprintf(stderr, "%s: Specified precision is too small\n", argv[0]);
        exit(1);
    }

    if (wellwidth <= 0.0) {
        fprintf(stderr, "%s: Specified well width is too small\n", argv[0]);
        exit(1);
    }

    if (basesize < N) {
        fprintf(stderr, "%s: Specified base size is too small\n", argv[0]);
        exit(1);
    }
    if(is_number(pot) && strtod(pot, NULL) != 0.0) {


        fprintf(stderr, "%s: Warning - You are using a constant potential that is not 0.0. This may cause the numerical routines to fail!\n", argv[0]);
    }
}

typedef struct {
    char* potential_string;
    char* filename;
    int calc_coeffecients;
    int output_potential;
} program_input;

program_input parse_cmd(int argc, char* argv[]) {
    if(argc == 6 && !strcmp(argv[1], "what") && !strcmp(argv[2], "is") && !strcmp(argv[3], "best") && !strcmp(argv[4], "in") && (!strcmp(argv[5], "life") || !strcmp(argv[5], "life?"))) {
        printf("Crush your enemies. See them driven before you. Hear the lamentations of their women.\n");
        exit(0);
    }
    program_input r;
    r.potential_string = NULL;
    r.filename = NULL;
    r.calc_coeffecients = 1;
    r.output_potential = 0;

    int c;

    while (1)
    {
        struct option long_options[] =
                {
                        {"help",                  no_argument, 0, 'h'},
                        {"potential",             required_argument, 0, 'V'},
                        {"filename",              required_argument, 0, 'o'},
                        {"basesize",              required_argument, 0, 'b'},
                        {"wellwidth",             required_argument, 0, 'L'},
                        {"number-of-par",        required_argument, 0, 'N'},
                        {"precision",             required_argument, 0, 'p'},
                        {"just-energy",           no_argument, &r.calc_coeffecients, 'E'},
                        {"just-potential",        no_argument, &r.output_potential, 'D'},
                        {"abs-solver",            required_argument, 0, 'l'},
                        {"rel-solver",            required_argument, 0, 'k'},
                        {"abs-final",             required_argument, 0, 'j'},
                        {"rel-final",             required_argument, 0, 'f'},
                        {0, 0, 0, 0}
                };
        /* getopt_long stores the option index here. */
        int option_index = 0;

        c = getopt_long (argc, argv, "hV:o:b:L:N:p:DEl:k:j:f:",
                         long_options, &option_index);

        /* Detect the end of the options. */
        if (c == -1)
            break;

        switch (c)
        {
            case 0:
                /* If this option set a flag, do nothing else now. */
                if (long_options[option_index].flag != 0)
                    break;
                printf ("option %s", long_options[option_index].name);
                if (optarg)
                    printf (" with arg %s", optarg);
                printf ("\n");
                break;
            case 'V':
                r.potential_string = optarg;
                break;
            case 'o':
                r.filename = optarg;
                break;
            case 'b':
                basesize = atoi(optarg);
                break;
            case 'L':
                wellwidth = strtod(optarg, NULL);
                break;
            case 'N':
                N = atoi(optarg);
                break;
            case 'p':
                prec = atoi(optarg);
                break;
            case 'l':
                int_prec_abs_Solver = strtod(optarg, NULL);
                break;
            case 'k':
                int_prec_rel_Solver = strtod(optarg, NULL);
                break;
            case 'j':
                int_prec_abs_final = strtod(optarg, NULL);
                break;
            case 'f':
                int_prec_rel_final = strtod(optarg, NULL);
                break;
            case 'h':
                printLongUsage();
                exit(0);
            case 'D':
                r.output_potential = 1;
                break;
            case 'E':
                r.calc_coeffecients = 0;
                break;
            default:
                abort ();
        }
    }

    return r;
}

int main(int argc, char* argv[])
{
    program_input in = parse_cmd(argc, argv);
    FILE* output = stdout;
    sanity_check(argv, N, prec, wellwidth, basesize, in.potential_string);

    if(in.filename) output = fopen(in.filename, "a");
    add_variable_builtin("N", (double)N);
    add_variable_builtins();

    // Parse input variables
    char** variables = (char**) malloc((argc-optind)*sizeof(char*));
    double* values = (double*) malloc((argc-optind)*sizeof(double));

    int single_value_parameter = 0;
    int multiple_value_parameters = 0;
    input_parameter* inputs = (input_parameter*) malloc((argc-optind)*sizeof(input_parameter));

    for (int i = optind; i < argc; i++ ) {
        input_parameter *ip = inputs+i-optind;
        parse_parameter(argv[i], inputs+i-optind);

        variables[i-optind] = strdup(ip->name);

        if(ip->n == 1) {
            add_variable(strdup(ip->name), ip->array[0]);
            values[i-optind] = ip->array[0];
            single_value_parameter++;
        }
        else if(ip->n > 1){
            inputs[multiple_value_parameters] = *ip;
            multiple_value_parameters++;
        }
    }

    // If we have multivalue parameters we need to iterate all, otherwise only one
    int number_of_runs = multiple_value_parameters > 0 ? inputs[0].n : 1;
    int equal_length = 1;
    for(int i=1; i < multiple_value_parameters; i++) {
        if(number_of_runs != inputs[i].n) {
            equal_length = 0;
        }
    }
    if(!equal_length) {
        fprintf(stderr, "Input parameters should be of equal length!\n");
        exit(1);
    }

    // Output header
    output_header(output, in.potential_string, variables, multiple_value_parameters + single_value_parameter);

    omp_set_num_threads(N+5);

    printf("Number of runs to be performed: %i\n", number_of_runs);
    printf("Starting calculation!\n");

    // Run main loop
    struct result r;
    time_t start, diff = 0;
    for(int i=0; i < number_of_runs; i++) {
        printf("Calculating %i/%i           \n", i + 1, number_of_runs);
        for (int j = 0; j < multiple_value_parameters; j++) {
            add_variable(inputs[j].name, inputs[j].array[i]);
            values[single_value_parameter + j] = inputs[j].array[i];
        }
        start = time(NULL);
        potential = compile(in.potential_string);
        if(in.output_potential)
            output_potential(variables, values, multiple_value_parameters+single_value_parameter, i);
        else {
            doSolve(N, &r, in.calc_coeffecients);
            diff += time(NULL) - start;

            output_default(output, &r, values, multiple_value_parameters + single_value_parameter);
        }

    }

    printf("Done!\n");
    printf("Total time elapsed: %i seconds\n", (int) diff);
    printf("Average time per run: %.1f seconds\n", (double) diff / (double) number_of_runs);

    if(in.filename) {
        printf("Result written to file: \"%s\"\n", in.filename);
        fclose(output);
    }

    return 0;
}
