//-------------------------------------------------------
// Coefficients of One-dimensional N-Atom Network (CONAN)
//-------------------------------------------------------

//    Copyright (C) 2016  The Few-Body Group, Department for Physics
//    and Astronomy, Aarhus University, Denmark

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.


//
// Created by jesper on 1/25/16.
//

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "cvector.h"

int prec = 256;

//-------------------------------
// Basic Vector Library
//-------------------------------

void vectorinit(vector* v, int dimension)
{
    v->dim = dimension;
    mpfr_t* tmp_data = (mpfr_t*) malloc(dimension * sizeof(mpfr_t));
    for (int i = 0; i < dimension; i++)
        mpfr_init2(tmp_data[i], prec);
    v->data = tmp_data;
}

void vectorset(vector v, int entry, mpfr_t value)
{
    assert(entry < v.dim);
    mpfr_set(v.data[entry], value, MPFR_RNDN);
}

void vectorset_d(vector v, int entry, double value)
{
    assert(entry < v.dim);
    mpfr_set_d(v.data[entry], value, MPFR_RNDN);
}

void vectorget(vector v, int entry, mpfr_t output)
{
    assert(entry < v.dim);
    mpfr_set(output, v.data[entry], MPFR_RNDN);
}

void vectorfree(vector* v)
{
    for (int i = 0; i < v->dim; i++)
        mpfr_clear(v->data[i]);
    free(v->data);
    v->data = NULL;
}

void vectorscale(vector v, mpfr_t scale, vector result)
{
    assert(v.dim == result.dim);
    mpfr_t tmp;
    mpfr_init2(tmp, prec);
    for (int i = 0; i < v.dim; i++) {
        vectorget(v, i, tmp);
        mpfr_mul(tmp, tmp, scale, MPFR_RNDN);
        vectorset(result, i, tmp);
    }
    mpfr_clear(tmp);
}

void vectorscale_d(vector v, double scale, vector result)
{
    assert(v.dim == result.dim);
    mpfr_t tmp;
    mpfr_init2(tmp, prec);
    for (int i = 0; i < v.dim; i++) {
        vectorget(v, i, tmp);
        mpfr_mul_d(tmp, tmp, scale, MPFR_RNDN);
        vectorset(result, i, tmp);
    }
    mpfr_clear(tmp);
}

// Stores v-w in result.
void vectorsub(vector v, vector w, vector result)
{
    assert(v.dim == w.dim);
    mpfr_t tmp1, tmp2;
    mpfr_inits2(prec, tmp1, tmp2, (void*) NULL);
    for (int i = 0; i < v.dim; i++) {
        vectorget(v, i, tmp1);
        vectorget(w, i, tmp2);
        mpfr_sub(tmp1, tmp1, tmp2, MPFR_RNDN);
        vectorset(result, i, tmp1);
    }
    mpfr_clears(tmp1, tmp2, (void*) NULL);
}

void vectoradd(vector v, vector w, vector result)
{
    assert(v.dim == w.dim);
    mpfr_t tmp1, tmp2;
    mpfr_inits2(prec, tmp1, tmp2, (void*) NULL);
    for (int i = 0; i < v.dim; i++) {
        vectorget(v, i, tmp1);
        vectorget(w, i, tmp2);
        mpfr_add(tmp1, tmp1, tmp2, MPFR_RNDN);
        vectorset(result, i, tmp1);
    }
    mpfr_clears(tmp1, tmp2, (void*) NULL);
}

// Shifts every entry in the vector by +1.
void vectorshift(vector v, vector result)
{
    assert(v.dim = result.dim);
    mpfr_t tmp;
    mpfr_init2(tmp, prec);
    for (int i = v.dim - 2; i >= 0; i--) {
        vectorget(v, i, tmp);
        vectorset(result, i + 1, tmp);
    }
    vectorset_d(result, 0, 0);
    mpfr_clear(tmp);
}

void vectorset_all_d(vector v, double d)
{
    for (int i = 0; i < v.dim; i++)
        vectorset_d(v, i, d);
}

void vectordiv(vector v, vector w, vector result) // v/w entrywise;
{
    assert(v.dim == w.dim);
    assert(result.dim == v.dim);
    mpfr_t tmp_num, tmp_den, tmp_frac;
    mpfr_inits2(prec, tmp_num, tmp_den, tmp_frac, (void*) NULL);
    for (int i = 0; i < v.dim; i++) {
        vectorget(v, i, tmp_num);
        vectorget(w, i, tmp_den);
        mpfr_div(tmp_frac, tmp_num, tmp_den, MPFR_RNDN);
        vectorset(result, i, tmp_frac);
    }
    mpfr_clears(tmp_num, tmp_den, tmp_frac, (void*) NULL);
}

//-------------------------------
// Basic Matrix Library
//-------------------------------

void matrixinit(matrix* M, int rows, int columns)
{
    M->row = rows;
    M->col = columns;
    mpfr_t* tmp_data = malloc(rows * columns * sizeof(mpfr_t));
    for (int i = 0; i < rows * columns; i++)
        mpfr_init2(tmp_data[i], prec);
    M->data = tmp_data;
}

void matrixset(matrix M, int row, int column, mpfr_t value)
{
    assert(row < M.row);
    assert(column < M.col);
    mpfr_set(M.data[row * M.row + column], value, MPFR_RNDN);
}

void matrixset_d(matrix M, int row, int column, double value)
{
    assert(row < M.row);
    assert(column < M.col);
    mpfr_set_d(M.data[row * M.row + column], value, MPFR_RNDN);
}

void matrixget(matrix M, int row, int column, mpfr_t output)
{
    assert(row < M.row);
    assert(column < M.col);
    mpfr_set(output, M.data[row * M.row + column], MPFR_RNDN);
}

void matrixfree(matrix* M)
{
    for (int i = 0; i < M->row * M->col; i++)
        mpfr_clear(M->data[i]);
    free(M->data);
    M->data = NULL;
}

void matrixset_I(matrix M)
{
    mpfr_t tmp1, tmp0;
    mpfr_init2(tmp1, prec);
    mpfr_init2(tmp0, prec);
    mpfr_set_d(tmp1, 1.0, MPFR_RNDN);
    mpfr_set_d(tmp0, 0.0, MPFR_RNDN);

    for (int i = 0; i < M.row; i++) {
        for (int j = 0; j < M.col; j++) {
            if (i == j)
                matrixset(M, i, i, tmp1);
            else
                matrixset(M, i, j, tmp0);
        }
    }

    mpfr_clear(tmp0);
    mpfr_clear(tmp1);
}

void matrixset_Mat_Sym(gsl_matrix* A, matrix M)
{
    int dim = M.col;
    assert((int) A->size2 == dim);
    assert((int) A->size1 == dim);
    assert(M.row == dim);
    for (int n = 0; n < dim; n++) {
        for (int m = n + 1; m < dim; m++) {
            double tmp = gsl_matrix_get(A, n, m);
            matrixset_d(M, n, m, tmp);
            matrixset_d(M, m, n, tmp);
        }
        double tmp = gsl_matrix_get(A, n, n);
        matrixset_d(M, n, n, tmp);
    }
}

//-------------------------------
// Mixed Matrix/Vector functions
//-------------------------------

// Multiplies a vector by Row i, and takes the dotproduct with Row j.
double RowDotVecMulRow(vector v, matrix M, int j, int i)
{
    assert(v.dim == M.row);
    assert(j < M.col);
    assert(i < M.col);
    mpfr_t tmp1, tmp2, sum;
    mpfr_inits2(prec, tmp1, tmp2, sum, (void*) NULL);
    mpfr_set_d(sum, 0, MPFR_RNDN);
    for (int n = 0; n < v.dim; n++) {
        matrixget(M, i, n, tmp1);
        vectorget(v, n, tmp2);
        mpfr_mul(tmp1, tmp1, tmp2, MPFR_RNDN);
        matrixget(M, j, n, tmp2);
        mpfr_mul(tmp1, tmp1, tmp2, MPFR_RNDN);
        mpfr_add(sum, sum, tmp1, MPFR_RNDN);
    }
    double result = mpfr_get_d(sum, MPFR_RNDN);
    mpfr_clears(tmp1, tmp2, sum, (void*) NULL);
    return result;
}
