//-------------------------------------------------------
// Coefficients of One-dimensional N-Atom Network (CONAN)
//-------------------------------------------------------

//    Copyright (C) 2016  The Few-Body Group, Department for Physics
//    and Astronomy, Aarhus University, Denmark

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.


//
// Created by jesper on 1/25/16.
//

#include "expparser.h"

const size_t functions_count = sizeof(functions) / sizeof(functions[0]);
const size_t num_binary_operators = sizeof(binary_operators) / sizeof(binary_operators[0]);

struct expression* make_expr(enum expression_type type)
{
    struct expression* expr =
            (struct expression*) calloc(sizeof(struct expression), 1);
    if (!expr) {
        fprintf(stderr, "Out of memory\n");
        exit(1);
    }
    expr->type = type;
    return expr;
}

int is_special_character(char c)
{
    return c == '(' || c == ')' || c == '+' || c == '-' ||
           c == '*' || c == '/' || c == '^' || c == ',';
}

int lex(const char** stringp, struct token* tokenp)
{
    const char* string = *stringp;
    if (isspace((unsigned char) string[0]))
        string++;
    if (!string[0])
        return 0;
    if (is_special_character(string[0])) {
        *stringp = string + 1;
        *tokenp = (struct token) { string, 1 };
        return 1;
    }
    char* endptr;
    strtod(string, &endptr);
    if (endptr != string) {
        *stringp = endptr;
        *tokenp = (struct token) { string, endptr - string };
        return 1;
    }
    size_t result = 1;
    while (string[result] &&
           !isspace((unsigned char) string[result]) &&
           !is_special_character(string[result]))
        result++;
    *stringp = string + result;
    *tokenp = (struct token) { string, result };
    return 1;
}

int precedence(enum expression_type type)
{
    switch (type) {
        case EXPRESSION_TYPE_PARENTHESIS:
        case EXPRESSION_TYPE_FUNCTION_CALL:
        case EXPRESSION_TYPE_FUNCTION_PARAMS:
            return 0;
        case EXPRESSION_TYPE_ADDITION:
        case EXPRESSION_TYPE_SUBTRACTION:
            return 1;
        case EXPRESSION_TYPE_MULTIPLICATION:
        case EXPRESSION_TYPE_DIVISION:
            return 2;
        case EXPRESSION_TYPE_ADDITION_UNARY:
        case EXPRESSION_TYPE_SUBTRACTION_UNARY:
            return 3;
        case EXPRESSION_TYPE_POWER:
            return 4;
        case EXPRESSION_TYPE_NUMBER:
        case EXPRESSION_TYPE_IDENTIFIER:
            return 5;
    }
    return 0;
}

struct expression* parse(const char* string)
{
    struct expression* result = NULL;
    struct expression** partial_ptr = &result;
    struct expression* partial = NULL;
    struct expression** last_ptr = NULL;
    struct expression* last = NULL;
    struct token token;
    size_t parenthesis_level = 0;
    while (lex(&string, &token)) {
        struct expression* expr;
        assert(!result || !result->parent);
        assert(!result || result->parent_ptr == &result);
        string = token.string + token.length;
        if (!last && token.length == 1 && token.string[0] == '+') {
            expr = make_expr(EXPRESSION_TYPE_ADDITION_UNARY);
            expr->parent = partial;
            expr->parent_ptr = partial_ptr;
            *partial_ptr = partial = expr;
            partial = expr;
            partial_ptr = &expr->first;
            last = NULL;
            last_ptr = NULL;
            continue;
        }
        if (!last && token.length == 1 && token.string[0] == '-') {
            expr = make_expr(EXPRESSION_TYPE_SUBTRACTION_UNARY);
            expr->parent = partial;
            expr->parent_ptr = partial_ptr;
            *partial_ptr = partial = expr;
            partial = expr;
            partial_ptr = &expr->first;
            last = NULL;
            last_ptr = NULL;
            continue;
        }
        if (token.length == 1 && token.string[0] == '(') {
            if (last) {
                expr = make_expr(EXPRESSION_TYPE_FUNCTION_CALL);
                expr->first = last;
                expr->parent = last->parent;
                expr->parent_ptr = last->parent_ptr;
                last->parent = expr;
                last->parent_ptr = &expr->first;
                *last_ptr = last = expr;
                partial = expr;
                partial_ptr = &expr->second;
                last = NULL;
                last_ptr = NULL;
            } else {
                expr = make_expr(EXPRESSION_TYPE_PARENTHESIS);
                expr->parent = partial;
                expr->parent_ptr = partial_ptr;
                *partial_ptr = partial = expr;
                partial = expr;
                partial_ptr = &expr->first;
                last = NULL;
                last_ptr = NULL;
            }
            parenthesis_level++;
            continue;
        }
        if (token.length == 1 && token.string[0] == ',') {
            if ( !parenthesis_level )
            {
                fprintf(stderr, "Syntax error: Unexpected ',' token\n");
                exit(1);
            }
            if (!last) {
                fprintf(stderr, "Syntax error: Expected expression before ',' token\n");
                exit(1);
            }
            while (partial->parent &&
                   partial->type != EXPRESSION_TYPE_FUNCTION_PARAMS &&
                   partial->parent->type != EXPRESSION_TYPE_PARENTHESIS &&
                   partial->parent->type != EXPRESSION_TYPE_FUNCTION_CALL) {
                partial = partial->parent;
                partial_ptr = partial->parent_ptr;
            }
            expr = make_expr(EXPRESSION_TYPE_FUNCTION_PARAMS);
            expr->first = partial;
            expr->parent = partial->parent;
            expr->parent_ptr = partial->parent_ptr;
            partial->parent = expr;
            partial->parent_ptr = &expr->first;
            *partial_ptr = partial = expr;
            partial = expr;
            partial_ptr = &expr->second;
            last = NULL;
            last_ptr = NULL;
            continue;
        }
        if (token.length == 1 && token.string[0] == ')') {
            if ( !parenthesis_level )
            {
                fprintf(stderr, "Syntax error: Unexpected ')' token\n");
                exit(1);
            }
            if ( partial->type == EXPRESSION_TYPE_FUNCTION_PARAMS )
            {
                fprintf(stderr, "Syntax error: Expected expression after ',' token but before ')' token\n");
                exit(1);
            }
            if (!*partial_ptr) {
                fprintf(stderr, "Syntax error in expression before ')' token\n");
                exit(1);
            }
            while (partial->parent &&
                   partial->type != EXPRESSION_TYPE_PARENTHESIS &&
                   partial->type != EXPRESSION_TYPE_FUNCTION_CALL) {
                partial = partial->parent;
                partial_ptr = partial->parent_ptr;
            }
            last = partial;
            last_ptr = partial->parent_ptr;
            parenthesis_level--;
            continue;
        }
        int handled = 0;
        for (size_t i = 0; token.length == 1 && i < num_binary_operators; i++) {
            const struct binary_operator* binop = &binary_operators[i];
            if (token.string[0] != binop->c)
                continue;
            if (!last) {
                fprintf(stderr, "Expected expression before '%c' token\n", binop->c);
                exit(1);
            }
            while (partial->parent &&
                   precedence(binop->type) <= precedence(partial->parent->type)) {
                partial = partial->parent;
            }
            assert(!result || result->parent_ptr == &result);
            expr = make_expr(binop->type);
            expr->first = partial;
            expr->parent = partial->parent;
            expr->parent_ptr = partial->parent_ptr;
            *partial->parent_ptr = expr;
            partial->parent = expr;
            partial->parent_ptr = &expr->first;
            partial = expr;
            partial_ptr = &expr->second;
            last = NULL;
            last_ptr = NULL;
            handled = 1;
            break;
        }
        if (handled)
            continue;
        char* endptr;
        double value = value = strtod(token.string, &endptr);
        if (endptr != token.string) {
            expr = make_expr(EXPRESSION_TYPE_NUMBER);
            expr->value = value;
        } else {
            expr = make_expr(EXPRESSION_TYPE_IDENTIFIER);
            expr->token = token;
        }
        expr->parent = partial;
        expr->parent_ptr = last_ptr = partial_ptr;
        *partial_ptr = last = partial = expr;
    }
    if (parenthesis_level) {
        fprintf(stderr, "Syntax error: Expected ')' before expression end\n");
        exit(1);
    }
    if (!result) {
        fprintf(stderr, "Syntax error: Expression is empty\n");
        exit(1);
    }
    if (!*partial_ptr) {
        fprintf(stderr, "Syntax error in expression\n");
        exit(1);
    }
    result->parent = NULL;
    result->parent_ptr = NULL;
    return result;
}



static struct variable* variables = NULL;
static size_t variables_used = 0;
static size_t variables_allocated = 0;

struct variable* current_variables() {
    return variables;
}

int n_current_variables() {
    return variables_used;
}

struct expression* disambiguate(struct expression* expr)
{
    //struct expression* tmp;
    switch (expr->type) {
        case EXPRESSION_TYPE_NUMBER:
            break;
        case EXPRESSION_TYPE_IDENTIFIER:
            if (expr->token.length == 1) {
                if (expr->token.string[0] == 'x' ||
                    expr->token.string[0] == 'L' )
                    return expr;
            }
            for (size_t i = 0; i < variables_used; i++) {
                const char* name = variables[i].name;
                if (strncmp(name, expr->token.string, expr->token.length) != 0)
                    continue;
                if (name[expr->token.length] != '\0')
                    continue;
                expr->type = EXPRESSION_TYPE_NUMBER;
                expr->value = variables[i].value;
                return expr;
            }
            for (size_t i = 0; !expr->function && i < functions_count; i++) {
                const struct function* func = &functions[i];
                if (strlen(func->name) != expr->token.length)
                    continue;
                if (memcmp(func->name, expr->token.string, expr->token.length) != 0)
                    continue;
                return expr;
            }
            fprintf(stderr, "Expression used variable named '");
            fwrite(expr->token.string, 1, expr->token.length, stderr);
            fprintf(stderr, "' but it hasn't been declared\n");
            exit(1);
        case EXPRESSION_TYPE_ADDITION_UNARY:
        case EXPRESSION_TYPE_SUBTRACTION_UNARY:
            expr->first = disambiguate(expr->first);
            break;
        case EXPRESSION_TYPE_ADDITION:
        case EXPRESSION_TYPE_SUBTRACTION:
        case EXPRESSION_TYPE_MULTIPLICATION:
        case EXPRESSION_TYPE_DIVISION:
        case EXPRESSION_TYPE_POWER:
        case EXPRESSION_TYPE_FUNCTION_PARAMS:
            expr->first = disambiguate(expr->first);
            expr->second = disambiguate(expr->second);
            break;
        case EXPRESSION_TYPE_PARENTHESIS:
            expr->first = disambiguate(expr->first);
            //tmp = expr->first;
            //free(expr);
            //expr = tmp;
            break;
        case EXPRESSION_TYPE_FUNCTION_CALL:
            expr->first = disambiguate(expr->first);
            expr->second = disambiguate(expr->second);
            if (expr->first->type != EXPRESSION_TYPE_IDENTIFIER) {
                fprintf(stderr, "Invalid function call on non-identifier\n");
                exit(1);
            }
            struct token name_token = expr->first->token;
            for (size_t i = 0; !expr->function && i < functions_count; i++) {
                const struct function* func = &functions[i];
                if (strlen(func->name) != name_token.length)
                    continue;
                if (memcmp(func->name, name_token.string, name_token.length) != 0)
                    continue;
                expr->function = func;
            }
            if (!expr->function) {
                fprintf(stderr, "Unknown function '");
                fwrite(name_token.string, 1, name_token.length, stderr);
                fprintf(stderr, "'\n");
                exit(1);
            }
            break;
    }
    return expr;
}

struct expression* compile(const char* input)
{

    // Since we can not handle two consecutive end-paranthesis, "))", we add
    // some "*1"'s so )) -> )*1):
    size_t length = strlen(input)+1;

    //new string is three times the size - corresponding to worst case, i.e everything is end paranthesis
    char *new_string = (char*) malloc(3*length*sizeof(char));

    int j=0;
    for(int i=0; i<length; i++) {
        if(input[i] == ')'&& input[i+1] == ')') {
            new_string[j] = ')';
            new_string[j+1] = '*';
            new_string[j+2] = '1';
            j+=3;
        }
        else {
            new_string[j] = input[i];
            j++;
        }
    }
    new_string[j] = '\0';

    return disambiguate(parse(new_string));
}

double eval(struct expression* expr, double x, double L)
{
    switch (expr->type) {
        case EXPRESSION_TYPE_NUMBER:
            return expr->value;
        case EXPRESSION_TYPE_IDENTIFIER:
            if (expr->token.length == 1 && expr->token.string[0] == 'x')
                return x;
            if (expr->token.length == 1 && expr->token.string[0] == 'L')
                return L;
            return 0.0;
        case EXPRESSION_TYPE_ADDITION_UNARY:
            return +eval(expr->first, x, L);
        case EXPRESSION_TYPE_SUBTRACTION_UNARY:
            return -eval(expr->first, x, L);
        case EXPRESSION_TYPE_ADDITION:
            return eval(expr->first, x, L) + eval(expr->second, x, L);
        case EXPRESSION_TYPE_SUBTRACTION:
            return eval(expr->first, x, L) - eval(expr->second, x, L);
        case EXPRESSION_TYPE_MULTIPLICATION:
            return eval(expr->first, x, L) * eval(expr->second, x, L);
        case EXPRESSION_TYPE_DIVISION:
            return eval(expr->first, x, L) / eval(expr->second, x, L);
        case EXPRESSION_TYPE_POWER:
            return pow(eval(expr->first, x, L), eval(expr->second, x, L));
        case EXPRESSION_TYPE_PARENTHESIS:
            return eval(expr->first, x, L);
        case EXPRESSION_TYPE_FUNCTION_CALL:
            assert(expr->function);
            switch (expr->function->type) {
                case FUNCTION_TYPE_D:
                {
                    double a = eval(expr->second, x, L);
                    return expr->function->u.f_d(a);
                } break;
                case FUNCTION_TYPE_DD:
                {
                    double a = eval(expr->second->first, x, L);
                    double b = eval(expr->second->second, x, L);
                    return expr->function->u.f_dd(a, b);
                } break;
                case FUNCTION_TYPE_DDD:
                {
                    double a = eval(expr->second->first, x, L);
                    double b = eval(expr->second->second->first, x, L);
                    double c = eval(expr->second->second->second, x, L);
                    return expr->function->u.f_ddd(a, b, c);
                } break;
            }
        case EXPRESSION_TYPE_FUNCTION_PARAMS:
            break;
    }
    return 0.0;
}

void add_variable(char* name, double value)
{
    for (size_t i = 0; i < variables_used; i++) {
        if (!strcmp(variables[i].name, name)) {
            variables[i].value = value;
//            free(name);
            return;
        }
    }
    if (variables_used == variables_allocated) {
        size_t new_allocated = 2 * variables_allocated;
        if (new_allocated == 0)
            new_allocated = 16;
        size_t new_size = new_allocated * sizeof(struct variable);
        variables = (struct variable*)realloc(variables, new_size);
        if (!variables) {
            fprintf(stderr, "Out of memory\n");
            exit(1);
        }
        variables_allocated = new_allocated;
    }
    struct variable* var = &variables[variables_used++];
    var->name = name;
    var->value = value;
}

void add_variable_builtin(const char* name, double value)
{
    char* name_copy = strdup(name);
    if (!name_copy) {
        fprintf(stderr, "Out of memory\n");
        exit(1);
    }
    add_variable(name_copy, value);
}

void add_variable_builtins()
{
    add_variable_builtin("e", M_E);
    add_variable_builtin("pi", M_PI);
}
