//-------------------------------------------------------
// Coefficients of One-dimensional N-Atom Network (CONAN)
//-------------------------------------------------------

//    Copyright (C) 2016  The Few-Body Group, Department for Physics
//    and Astronomy, Aarhus University, Denmark

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.


//
// Created by jesper on 1/25/16.
//

#ifndef CONAN_EXPPARSER_H
#define CONAN_EXPPARSER_H

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <omp.h>


//-------------------------------
// Variables
//-------------------------------


//-------------------------------
// Enums
//-------------------------------

enum expression_type {
    EXPRESSION_TYPE_NUMBER,
    EXPRESSION_TYPE_IDENTIFIER,
    EXPRESSION_TYPE_ADDITION_UNARY,
    EXPRESSION_TYPE_SUBTRACTION_UNARY,
    EXPRESSION_TYPE_ADDITION,
    EXPRESSION_TYPE_SUBTRACTION,
    EXPRESSION_TYPE_MULTIPLICATION,
    EXPRESSION_TYPE_DIVISION,
    EXPRESSION_TYPE_POWER,
    EXPRESSION_TYPE_PARENTHESIS,
    EXPRESSION_TYPE_FUNCTION_CALL,
    EXPRESSION_TYPE_FUNCTION_PARAMS,
};

enum function_type {
    FUNCTION_TYPE_D,
    FUNCTION_TYPE_DD,
    FUNCTION_TYPE_DDD,
};

//-------------------------------
// Unions
//-------------------------------

union function_union {
    double (*f_d)(double);
    double (*f_dd)(double, double);
    double (*f_ddd)(double, double, double);
};

//-------------------------------
// Structs
//-------------------------------

// User-assigned variables.
struct variable
{
    char* name;
    double value;
};

struct token {
    const char* string;
    size_t length;
};

struct expression {
    enum expression_type type;
    struct expression* first;
    struct expression* second;
    struct expression* parent;
    struct expression** parent_ptr;
    double value;
    struct token token;
    const struct function* function;
};

struct binary_operator {
    char c;
    enum expression_type type;
};

struct function {
    const char* name;
    enum function_type type;
    union function_union u;
};


//-------------------------------
// Methods
//-------------------------------

struct expression* make_expr(enum expression_type type);

int is_special_character(char c);

int lex(const char** stringp, struct token* tokenp);

int precedence(enum expression_type type);

struct expression* parse(const char* string);

struct expression* disambiguate(struct expression* expr);

struct expression* compile(const char* input);

double eval(struct expression* expr, double x, double L);

struct variable* current_variables();

int n_current_variables();

//-------------------------------
// Arrays
//-------------------------------


static const struct binary_operator binary_operators[] = {
        { '+', EXPRESSION_TYPE_ADDITION },
        { '-', EXPRESSION_TYPE_SUBTRACTION },
        { '*', EXPRESSION_TYPE_MULTIPLICATION },
        { '/', EXPRESSION_TYPE_DIVISION },
        { '^', EXPRESSION_TYPE_POWER },
};
extern const size_t num_binary_operators;



static const struct function functions[] = {
        { "acos", FUNCTION_TYPE_D, { .f_d = acos } },
        { "acosh", FUNCTION_TYPE_D, { .f_d = acosh } },
        { "asin", FUNCTION_TYPE_D, { .f_d = asin } },
        { "asinh", FUNCTION_TYPE_D, { .f_d = asinh } },
        { "atan2", FUNCTION_TYPE_DD, { .f_dd = atan2 } },
        { "atan", FUNCTION_TYPE_D, { .f_d = atan } },
        { "atanh", FUNCTION_TYPE_D, { .f_d = atanh } },
        { "cbrt", FUNCTION_TYPE_D, { .f_d = cbrt } },
        { "ceil", FUNCTION_TYPE_D, { .f_d = ceil } },
        { "copysign", FUNCTION_TYPE_DD, { .f_dd = copysign } },
        { "cos", FUNCTION_TYPE_D, { .f_d = cos } },
        { "cosh", FUNCTION_TYPE_D, { .f_d = cosh } },
        { "erfc", FUNCTION_TYPE_D, { .f_d = erfc } },
        { "erf", FUNCTION_TYPE_D, { .f_d = erf } },
        { "exp2", FUNCTION_TYPE_D, { .f_d = exp2 } },
        { "exp", FUNCTION_TYPE_D, { .f_d = exp } },
        { "expm1", FUNCTION_TYPE_D, { .f_d = expm1 } },
        { "fabs", FUNCTION_TYPE_D, { .f_d = fabs } },
        { "fdim", FUNCTION_TYPE_DD, { .f_dd = fdim } },
        { "floor", FUNCTION_TYPE_D, { .f_d = floor } },
        { "fma", FUNCTION_TYPE_DDD, { .f_ddd = fma } },
        { "fmax", FUNCTION_TYPE_DD, { .f_dd = fmax } },
        { "fmin", FUNCTION_TYPE_DD, { .f_dd = fmin } },
        { "fmod", FUNCTION_TYPE_DD, { .f_dd = fmod } },
        { "hypot", FUNCTION_TYPE_DD, { .f_dd = hypot } },
        { "lgamma", FUNCTION_TYPE_D, { .f_d = lgamma } },
        { "log10", FUNCTION_TYPE_D, { .f_d = log10 } },
        { "log1p", FUNCTION_TYPE_D, { .f_d = log1p } },
        { "log2", FUNCTION_TYPE_D, { .f_d = log2 } },
        { "logb", FUNCTION_TYPE_D, { .f_d = logb } },
        { "log", FUNCTION_TYPE_D, { .f_d = log } },
        { "nearbyint", FUNCTION_TYPE_D, { .f_d = nearbyint } },
        { "nextafter", FUNCTION_TYPE_DD, { .f_dd = nextafter } },
        { "pow", FUNCTION_TYPE_DD, { .f_dd = pow } },
        { "remainder", FUNCTION_TYPE_DD, { .f_dd = remainder } },
        { "rint", FUNCTION_TYPE_D, { .f_d = rint } },
        { "round", FUNCTION_TYPE_D, { .f_d = round } },
        { "sin", FUNCTION_TYPE_D, { .f_d = sin } },
        { "sinh", FUNCTION_TYPE_D, { .f_d = sinh } },
        { "sqrt", FUNCTION_TYPE_D, { .f_d = sqrt } },
        { "tan", FUNCTION_TYPE_D, { .f_d = tan } },
        { "tanh", FUNCTION_TYPE_D, { .f_d = tanh } },
        { "tgamma", FUNCTION_TYPE_D, { .f_d = tgamma } },
        { "trunc", FUNCTION_TYPE_D, { .f_d = trunc } },
};
extern const size_t functions_count;

//-------------------------------
// Adding user-assigned variables
//-------------------------------

void add_variable(char* name, double value);

void add_variable_builtin(const char* name, double value);

void add_variable_builtins();


#ifdef DEBUG_EXPRESSION
static void pretty_inner(struct expression* expr, int inner)
{
	if (!expr) {
		fputs("<null>", stdout);
		return;
	}

	char c = '?';
	switch (expr->type) {
	case EXPRESSION_TYPE_NUMBER:
		printf("%.1f", expr->value);
		return;
	case EXPRESSION_TYPE_IDENTIFIER:
		fwrite(expr->token.string, 1, expr->token.length, stdout);
		return;
	case EXPRESSION_TYPE_ADDITION_UNARY:
		putchar('+');
		pretty_inner(expr->first, 1);
		return;
	case EXPRESSION_TYPE_SUBTRACTION_UNARY:
		putchar('-');
		pretty_inner(expr->first, 1);
		return;
	case EXPRESSION_TYPE_ADDITION: c = '+'; break;
	case EXPRESSION_TYPE_SUBTRACTION: c = '-'; break;
	case EXPRESSION_TYPE_MULTIPLICATION: c = '*'; break;
	case EXPRESSION_TYPE_DIVISION: c = '/'; break;
	case EXPRESSION_TYPE_POWER: c = '^'; break;
	case EXPRESSION_TYPE_PARENTHESIS:
		putchar('(');
		pretty_inner(expr->first, 1);
		putchar(')');
		return;
	case EXPRESSION_TYPE_FUNCTION_CALL:
		pretty_inner(expr->first, 1);
		putchar('(');
		if (expr->second)
			pretty_inner(expr->second, 0);
		putchar(')');
		return;
	case EXPRESSION_TYPE_FUNCTION_PARAMS:
		while (expr && expr->type == EXPRESSION_TYPE_FUNCTION_PARAMS) {
			pretty_inner(expr->first, inner);
			expr = expr->second;
			putchar(',');
			putchar(' ');
		}
		pretty_inner(expr, inner);
		return;
	}
	if (inner)
		putchar('(');
	if (expr->first && expr->first->parent != expr)
		fputs("<bad parent>", stdout);
	if (expr->first && expr->first->parent_ptr != &expr->first)
		fputs("<bad parent_ptr>", stdout);
	pretty_inner(expr->first, 1);
	printf(" %c ", c);
	if (expr->second && expr->second->parent != expr)
		fputs("<bad parent>", stdout);
	if (expr->second && expr->second->parent_ptr != &expr->second)
		fputs("<bad parent_ptr>", stdout);
	pretty_inner(expr->second, 1);
	if (inner)
		putchar(')');
}

static void pretty(struct expression* expr)
{
	pretty_inner(expr, 0);
}
#endif

#endif //CONAN_EXPPARSER_H
