//-------------------------------------------------------
// Coefficients of One-dimensional N-Atom Network (CONAN)
//-------------------------------------------------------

//    Copyright (C) 2016  The Few-Body Group, Department for Physics
//    and Astronomy, Aarhus University, Denmark

//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.


//
// Created by jesper on 1/31/16.
//

#include "physics.h"
#include <assert.h>
#include <math.h>
#include "expparser.h"

#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_integration.h>



double wellwidth = 100;
int N = -1; // marks error
int basesize = 300;
// Precision for Quantum-solver integrations.
double int_prec_abs_Solver = 1e-9;
double int_prec_rel_Solver = 1e-9;

// Precision of the final integrations
double int_prec_abs_final = 1e-5;
double int_prec_rel_final = 1e-5;

struct expression* potential;

double V(double x)
{
    return eval(potential, x, wellwidth);
}

//-------------------------------
// Dmitri's Jacobi-Diagonalization-Algorithm
//-------------------------------

void phi_func(mpfr_t app, mpfr_t aqq, mpfr_t apq, mpfr_t result)
{
    mpfr_t tmp1, tmp2, tmp3;
    mpfr_inits2(prec, tmp1, tmp2, tmp3, (void*) NULL);
    mpfr_sub(tmp1, aqq, app, MPFR_RNDN);
    mpfr_mul_d(tmp2, apq, 2.0, MPFR_RNDN);
    mpfr_atan2(tmp3, tmp2, tmp1, MPFR_RNDN);
    mpfr_mul_d(result, tmp3, 0.5, MPFR_RNDN);
    mpfr_clears(tmp1, tmp2, tmp3, (void*) NULL);
}

void app1_func(mpfr_t c, mpfr_t s, mpfr_t app, mpfr_t apq, mpfr_t aqq,
               mpfr_t result)
{
    mpfr_t tmp1, tmp2, tmp3;
    mpfr_inits2(prec, tmp1, tmp2, tmp3, (void*) NULL);
    mpfr_mul(tmp1, c, c, MPFR_RNDN);
    mpfr_mul(tmp1, tmp1, app, MPFR_RNDN);
    mpfr_mul(tmp2, apq, c, MPFR_RNDN);
    mpfr_mul(tmp2, tmp2, s, MPFR_RNDN);
    mpfr_mul_d(tmp2, tmp2, 2.0, MPFR_RNDN);
    mpfr_mul(tmp3, s, aqq, MPFR_RNDN);
    mpfr_mul(tmp3, tmp3, s, MPFR_RNDN);
    mpfr_sub(tmp1, tmp1, tmp2, MPFR_RNDN);
    mpfr_add(result, tmp1, tmp3, MPFR_RNDN);
    mpfr_clears(tmp1, tmp2, tmp3, (void*) NULL);
}

void aqq1_func(mpfr_t c, mpfr_t s, mpfr_t app, mpfr_t apq, mpfr_t aqq,
               mpfr_t result)
{
    mpfr_t tmp1, tmp2, tmp3;
    mpfr_inits2(prec, tmp1, tmp2, tmp3, (void*) NULL);
    mpfr_mul(tmp1, s, s, MPFR_RNDN);
    mpfr_mul(tmp1, tmp1, app, MPFR_RNDN);
    mpfr_mul(tmp2, apq, c, MPFR_RNDN);
    mpfr_mul(tmp2, tmp2, s, MPFR_RNDN);
    mpfr_mul_d(tmp2, tmp2, 2.0, MPFR_RNDN);
    mpfr_mul(tmp3, c, aqq, MPFR_RNDN);
    mpfr_mul(tmp3, tmp3, c, MPFR_RNDN);
    mpfr_add(tmp1, tmp1, tmp2, MPFR_RNDN);
    mpfr_add(result, tmp1, tmp3, MPFR_RNDN);
    mpfr_clears(tmp1, tmp2, tmp3, (void*) NULL);
}

void cx_minus_sy(mpfr_t x, mpfr_t y, mpfr_t c, mpfr_t s, mpfr_t result)
{
    mpfr_t tmp1, tmp2;
    mpfr_inits2(prec, tmp1, tmp2, (void*) NULL);
    mpfr_mul(tmp1, c, x, MPFR_RNDN);
    mpfr_mul(tmp2, s, y, MPFR_RNDN);
    mpfr_sub(result, tmp1, tmp2, MPFR_RNDN);
    mpfr_clears(tmp1, tmp2, (void*) NULL);
}

void cx_plus_sy(mpfr_t x, mpfr_t y, mpfr_t c, mpfr_t s, mpfr_t result)
{
    mpfr_t tmp1, tmp2;
    mpfr_inits2(prec, tmp1, tmp2, (void*) NULL);
    mpfr_mul(tmp1, c, x, MPFR_RNDN);
    mpfr_mul(tmp2, s, y, MPFR_RNDN);
    mpfr_add(result, tmp1, tmp2, MPFR_RNDN);
    mpfr_clears(tmp1, tmp2, (void*) NULL);
}

// e = eigenvalues, V = eigenvectors, upper triangle of A is destroyed.
int Jacobi(matrix A, vector e, matrix U)
{
    int changed;
    int rotations = 0;
    int n = A.row;
    mpfr_t tmp;
    mpfr_init2(tmp, prec);
    for (int i = 0; i < n; i++) {
        matrixget(A, i, i, tmp);
        vectorset(e, i, tmp);
    }
    matrixset_I(U);
    mpfr_t app, aqq, apq, phi, c, s, app1, aqq1, zero, aip, api, aiq, aqi, vip,
            viq;
    mpfr_inits2(prec, app, aqq, apq, phi, c, s, app1, aqq1, zero, aip, api, aiq,
                aqi, vip, viq, (void*) NULL);
    mpfr_set_d(zero, 0, MPFR_RNDN);
    do {
        changed = 0;
        for (int p = 0; p < n; p++) {
            for (int q = p + 1; q < n; q++) {
                vectorget(e, p, app);
                vectorget(e, q, aqq);
                matrixget(A, p, q, apq);
                phi_func(app, aqq, apq, phi);
                mpfr_cos(c, phi, MPFR_RNDN);
                mpfr_sin(s, phi, MPFR_RNDN);
                app1_func(c, s, app, apq, aqq, app1);
                aqq1_func(c, s, app, apq, aqq, aqq1);
                if (mpfr_cmp(app1, app) || mpfr_cmp(aqq1, aqq)) {
                    changed = 1;
                    rotations++;
                    vectorset(e, p, app1);
                    vectorset(e, q, aqq1);
                    matrixset(A, p, q, zero);
                    //#pragma omp parallel private(i,aip,aiq,api,aqi,tmp)
                    {
                        //#pragma omp sections nowait
                        {
                            //#pragma omp section
                            for (int i = 0; i < p; i++) {
                                matrixget(A, i, p, aip);
                                matrixget(A, i, q, aiq);
                                cx_minus_sy(aip, aiq, c, s, tmp);
                                matrixset(A, i, p, tmp);
                                cx_plus_sy(aiq, aip, c, s, tmp);
                                matrixset(A, i, q, tmp);
                            }
                            //#pragma omp section
                            for (int i = p + 1; i < q; i++) {
                                matrixget(A, p, i, api);
                                matrixget(A, i, q, aiq);
                                cx_minus_sy(api, aiq, c, s, tmp);
                                matrixset(A, p, i, tmp);
                                cx_plus_sy(aiq, api, c, s, tmp);
                                matrixset(A, i, q, tmp);
                            }
                            //#pragma omp section
                            for (int i = q + 1; i < n; i++) {
                                matrixget(A, p, i, api);
                                matrixget(A, q, i, aqi);
                                cx_minus_sy(api, aqi, c, s, tmp);
                                matrixset(A, p, i, tmp);
                                cx_plus_sy(aqi, api, c, s, tmp);
                                matrixset(A, q, i, tmp);
                            }
                            //#pragma omp section
                            for (int i = 0; i < n; i++) {
                                matrixget(U, i, p, vip);
                                matrixget(U, i, q, viq);
                                cx_minus_sy(vip, viq, c, s, tmp);
                                matrixset(U, i, p, tmp);
                                cx_plus_sy(viq, vip, c, s, tmp);
                                matrixset(U, i, q, tmp);
                            }
                        }
                    }
                }
            }
        }
    } while (changed != 0);
    mpfr_clears(app, aqq, apq, phi, c, s, app1, aqq1, zero, aip, api, aiq, aqi,
                vip, viq, tmp, (void*) NULL);
    return rotations;
}

//-------------------------------
// Solver for the quantum system
//-------------------------------

double integrand_V(double t, void* params_ptr)
{
    intparams_V* params = (intparams_V*) params_ptr;
    int m = *params->m;
    int k = *params->k;
    return (2.0 / wellwidth) * sin(m * M_PI * t / wellwidth) *
           sin(k * M_PI * t / wellwidth) * V(t);
}

void Hamiltonian(gsl_matrix* H)
{
    assert((int) H->size2 == basesize);
    assert((int) H->size1 == basesize);
    int m, k;
    double result, error;
    double konst = M_PI * M_PI / (wellwidth * wellwidth);
    intparams_V params;
    params.m = &m;
    params.k = &k;
    gsl_function integrand = { &integrand_V, (void*)(&params) };
    gsl_integration_workspace* w = gsl_integration_workspace_alloc(1000);

    for (m = 1; m <= basesize; m++) {
        for (k = m; k <= basesize; k++) {
            gsl_integration_qag(&integrand, 0, wellwidth, int_prec_abs_Solver,
                                 int_prec_rel_Solver, 1000, 6, w, &result, &error);
            gsl_matrix_set(H, m - 1, k - 1, result);
            gsl_matrix_set(H, k - 1, m - 1, result);
        }
    }
    for (m = 1; m <= basesize; m++) {
        double tmp = gsl_matrix_get(H, m - 1, m - 1);
        gsl_matrix_set(H, m - 1, m - 1, m * m * konst + tmp);
    }
    gsl_integration_workspace_free(w);
}

void Solver(gsl_matrix* C, gsl_vector* E)
{
    assert((int) C->size2 == basesize);
    int dim = C->size1;
    assert((int) E->size >= dim);
    gsl_matrix* H = gsl_matrix_alloc(basesize, basesize);
    gsl_matrix* Solution = gsl_matrix_alloc(basesize, basesize);
    Hamiltonian(H);
    gsl_eigen_symmv_workspace* w = gsl_eigen_symmv_alloc(basesize);
    gsl_eigen_symmv(H, E, Solution, w);
    gsl_eigen_symmv_sort(E, Solution, GSL_EIGEN_SORT_VAL_ASC);
    for (int n = 0; n < dim; n++)
        for (int m = 0; m < basesize; m++)
            gsl_matrix_set(C, n, m, gsl_matrix_get(Solution, m, n));
    gsl_matrix_free(H);
    gsl_matrix_free(Solution);
    gsl_eigen_symmv_free(w);
}

double Psi(double x, int i, gsl_matrix* C)
{
    double sum = 0.0;
    double konst = M_PI / wellwidth;
    for (int n = 1; n <= basesize; n++)
        sum += gsl_matrix_get(C, i, (n - 1)) * sin(n * konst * x);
    return sum * sqrt(2.0 / wellwidth); // normalize
}

double DPsi(double x, int i, gsl_matrix* C)
{
    double sum = 0;
    double konst = M_PI / wellwidth;
    for (int n = 1; n <= basesize; n++)
        sum += gsl_matrix_get(C, i, n - 1) * n * konst * cos(n * konst * x);
    return sum * sqrt(2.0 / wellwidth); // normalize
}

//-------------------------------
// Generating the B(x)-matrix
//-------------------------------

void B_Gen(double x, gsl_matrix* C, matrix B)
{
    assert((int) C->size2 == basesize);
    int dim = B.col;
    double piinv = M_1_PI;
    double pixa = M_PI * x / wellwidth;
    gsl_matrix* f = gsl_matrix_alloc(basesize, basesize);
    gsl_matrix* tmpMat = gsl_matrix_alloc(basesize, dim);
    gsl_matrix* Btmp = gsl_matrix_alloc(dim, dim);
    for (int n = 1; n <= basesize; n++) {
        for (int m = n + 1; m <= basesize; m++) {
            double tmp = piinv * ((1.0 / (m - n)) * sin((m - n) * pixa) -
                                  (1.0 / (m + n)) * sin((m + n) * pixa));
            gsl_matrix_set(f, m - 1, n - 1, tmp);
            gsl_matrix_set(f, n - 1, m - 1, tmp);
        }
        double tmp = piinv * (pixa - (1.0 / (2.0 * n)) * sin(2 * n * pixa));
        gsl_matrix_set(f, n - 1, n - 1, tmp);
    }
    gsl_blas_dgemm(CblasNoTrans, CblasTrans, 1, f, C, 0, tmpMat);
    gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, C, tmpMat, 0, Btmp);

    matrixset_Mat_Sym(Btmp, B);

    gsl_matrix_free(f);
    gsl_matrix_free(tmpMat);
    gsl_matrix_free(Btmp);
}

//-------------------------------
// Functions for calculating the derivatives
//-------------------------------

void Polynomial_Coefficients(vector eigenval, vector result)
{
    assert(eigenval.dim == result.dim - 1);
    int dim = eigenval.dim;
    mpfr_t tmp;
    mpfr_init2(tmp, prec);
    vector vprime;
    vectorinit(&vprime, dim + 1);
    vectorget(eigenval, 0, tmp);
    vectorset(result, 0, tmp);
    vectorset_d(result, 1, -1);
    for (int i = 2; i < dim + 1; i++)
        vectorset_d(result, i, 0);
    for (int i = 1; i < dim; i++) {
        vectorshift(result, vprime);
        vectorget(eigenval, i, tmp);
        vectorscale(result, tmp, result);
        vectorsub(result, vprime, result);
    }
    vectorfree(&vprime);
    mpfr_clear(tmp);
}

void Deriv_Matrix(vector eigenval, vector p, int k, vector result)
{
    assert(eigenval.dim == result.dim);
    assert(p.dim == eigenval.dim + 1);
    vectorset_all_d(result, 0);
    int dim = eigenval.dim;
    vector tmp, pD;
    vectorinit(&tmp, dim);
    vectorinit(&pD, dim);
    mpfr_t tmp_p;
    mpfr_init2(tmp_p, prec);
    vector D[dim - k];
    for (int n = 0; n < dim - k; n++)
        vectorinit(&D[n], dim);
    vectorset_all_d(tmp, 1);
    vectordiv(tmp, eigenval, D[0]);
    for (int n = 1; n < dim - k; n++)
        vectordiv(D[n - 1], eigenval, D[n]);
    for (int l = 0; l <= dim - 1 - k; l++) {
        vectorset_all_d(tmp, 0);
        for (int n = 0; n <= l; n++) {
            // Remember: zero indexed in code, one-indexed in note.
            vectorget(p, l - n, tmp_p);
            vectorscale(D[n], tmp_p, pD);
            vectoradd(tmp, pD, tmp);
        }
        vectorscale_d(tmp, gsl_sf_choose(dim - l - 2, k - 1), tmp);
        vectoradd(tmp, result, result);
    }
    vectorfree(&tmp);
    vectorfree(&pD);
    mpfr_clear(tmp_p);
    for (int n = 0; n < dim - k; n++)
        vectorfree(&D[n]);
}

//-------------------------------
// Constructing the integrand
//-------------------------------

double Integrand(double x, void* params_ptr)
{
    intparams* params = (intparams*) params_ptr;
    int dim = params->dim;
    int k = params->k;
    gsl_vector* E = params->E;
    gsl_matrix* C = params->C;
    double sum = 0;
    double Vx = V(x);
    matrix B, UT;
    matrixinit(&B, dim, dim);
    matrixinit(&UT, dim, dim);
    vector eigenval, p, Deriv;
    vectorinit(&eigenval, dim);
    vectorinit(&p, dim + 1);
    vectorinit(&Deriv, dim);
    B_Gen(x, C, B);
    Jacobi(B, eigenval, UT);
    Polynomial_Coefficients(eigenval, p);
    Deriv_Matrix(eigenval, p, k, Deriv);
    for (int i = 0; i < dim; i++) {
        for (int j = i + 1; j < dim; j++) {
            sum += ((Vx - gsl_vector_get(E, i)) * Psi(x, i, C)
                    * DPsi(x, j, C)
                    + (Vx - gsl_vector_get(E, j)) * Psi(x, j, C)
                      * DPsi(x, i, C))
                   * RowDotVecMulRow(Deriv, UT, j, i);
        }
        sum += (Vx - gsl_vector_get(E, i)) * Psi(x, i, C)
               * DPsi(x, i, C)
               * RowDotVecMulRow(Deriv, UT, i, i);
    }
    matrixfree(&B);
    matrixfree(&UT);
    vectorfree(&eigenval);
    vectorfree(&p);
    vectorfree(&Deriv);
    return sum;
}

//-------------------------------
// Final Coefficients
//-------------------------------

double I(int dim, int k, gsl_vector* E, gsl_matrix* C)
{
    assert(k < dim);
    intparams params;
    params.dim = dim;
    params.k = k;
    params.E = E;
    params.C = C;
    double result, err;
    gsl_integration_workspace* w = gsl_integration_workspace_alloc(1000);
    // The integral is defined in a way gsl understands.
    gsl_function gsl_integrand = { &Integrand, (void*)(&params) };
    gsl_integration_qag(&gsl_integrand, 0, wellwidth, int_prec_abs_final,
                         int_prec_rel_final, 1000, 6, w, &result, &err);
    gsl_integration_workspace_free(w);
    // Add boundary term.
    double psisum = 0;
    for (int i = 0; i < dim; i++)
        psisum += pow(DPsi(wellwidth, i, C), 2);
    // If boundary term is non-zero.
    return 2 * pow(-1, dim - k) * result + psisum;
}

void free_result(struct result* r) {
    gsl_vector_free(r->Ik);
    gsl_vector_free(r->E);
}
